/*
 * $Source: /home/gjvc/rnd/apache/modules/mod_log_access/mod_log_access-CURRENT/RCS/log_timeutils.c,v $
 * $Id: log_timeutils.c,v 1.26 2006/12/28 22:27:15 gjvc Exp $
 */


/* httpd */
#include "httpd.h"
#include "http_log.h"
#include "util_time.h"

/* log_timeutils */
#include "log_timeutils.h"


static const char *RCS_ID = "$Id: log_timeutils.c,v 1.26 2006/12/28 22:27:15 gjvc Exp $";


/* ---- begin static data ---- */

const char *interval_names[] = { "second", "minute", "hour", "day", "week", "month", "year", "infinity" };

/* ---- end static data ---- */


/* ---- begin configuration data ---- */

/* America and Europe disagree on Sunday or Monday as first day of week -- weeks_start_on_mondays is set if a %U specifier is encountered */
int weeks_start_on_mondays = 0;

/* ---- end configuration data ---- */


/* ---- begin time functions ---- */

/*
 * log_timeutils_explode_localtime
 *
 * @param   *apr_time_exp_t
 * @param   apr_time_t
 * @returns apr_status_t
 */

apr_status_t log_timeutils_explode_localtime (apr_time_exp_t *tm, apr_time_t t) {
    return apr_time_exp_lt( tm, t );                    /* ap_explode_recent_localtime */
    /* return ap_explode_recent_localtime( tm, t ); */  /* uses pre-exploded structures, so faster */
}


/*
 * log_timeutils_determine_interval
 * 
 * Examine the log file name specifier for strftime conversion
 * specifiers and determine the interval between log files
 * Smallest interval allowed is per minute
 *
 * @param   const char*
 * @returns interval_t
 */

interval_t log_timeutils_determine_interval(const char *spec) {

    interval_t interval = INTERVAL_INFINITY;
    char ch;

    if ( spec == NULL ) {
        return INTERVAL_INFINITY;
    }

    while ((ch = *spec++) != 0) {
        if (ch == '%') {
            ch = *spec++;
            if (!ch)
                break;

            switch (ch) {
                case 'y':   /* two digit year */
                case 'Y':   /* four digit year */
                    if ( interval > INTERVAL_YEAR ) {
                        interval = INTERVAL_YEAR;
                    }
                    break;

                case 'b':   /* short month name */
                case 'B':   /* full month name */
                case 'h':   /* abbreviated month name (non-standard) */
                case 'm':   /* month as two digit number (with leading zero) */
                    if ( interval > INTERVAL_MONTH ) {
                        interval = INTERVAL_MONTH;
                    }
                    break;

                case 'U':   /* week number (weeks start on Sunday) */
                case 'W':   /* week number (weeks start on Monday) */
                    if ( interval > INTERVAL_WEEK ) {
                        interval = INTERVAL_WEEK;
                        weeks_start_on_mondays = (ch == 'W');
                    }
                    break;

                case 'a':   /* short weekday name */
                case 'A':   /* full weekday name */
                case 'd':   /* day of the month (with leading zero) */
                case 'D':   /* full date spec (non-standard) */
                case 'e':   /* day of the month (with leading space -- non-standard) */
                case 'j':   /* day of the year (with leading zeroes) */
                case 'w':   /* day of the week (0-6) */
                case 'x':   /* full date spec */
                    if ( interval > INTERVAL_DAY ) {
                        interval = INTERVAL_DAY;
                    }
                    break;

                case 'H':   /* hour (24 hour clock) */
                case 'I':   /* hour (12 hour clock) */
                case 'p':   /* AM/PM indicator */
                    if ( interval > INTERVAL_HOUR ) {
                        interval = INTERVAL_HOUR;
                    }
                    break;

                case 'M':   /* minute */
                    if ( interval > INTERVAL_MINUTE ) {
                        interval = INTERVAL_MINUTE;
                    }
                    break;

                case 'c':   /* full time and date spec */
                case 'r':   /* full time spec (non-standard) */
                case 'R':   /* full time spec (non-standard) */
                case 's':   /* seconds since the epoch (GNU non-standard) */
                case 'S':   /* second */
                case 'T':   /* full time spec */
                    interval = INTERVAL_SECOND;
                    break;

                default:    /* ignore anything else */
                    break;
            }
        }
    }

    /* */
    return interval;
}


/*
 * log_timeutils_start_of_current_interval
 * 
 * Determine the time of the start of the interval containing a given time.
 * Break down the time with localtime(3) and subtract the number of seconds
 * since the start of the interval.
 * 
 * If the length of interval is equal to or longer than a day then check the
 * calculation is not thrown out by the start or end of daylight saving time.
 *
 * @param       apr_time_t
 * @param       interval_t
 * @returns     apr_time_t
 */

apr_time_t log_timeutils_start_of_current_interval(apr_time_t start_time, interval_t interval) {

    apr_time_exp_t  tm_initial, tm_adjusted;
    int             expected_mday = -1;

    /* */
    log_timeutils_explode_localtime( &tm_initial, start_time );

    /* */
    switch (interval) {
        case INTERVAL_INFINITY:
            start_time = 0;
            break;

        case INTERVAL_YEAR:
        case INTERVAL_MONTH:
        case INTERVAL_WEEK:
        case INTERVAL_DAY:
            switch (interval) {
                case INTERVAL_INFINITY:
                    start_time = 0;
                    break;

                case INTERVAL_YEAR:
                    start_time -= apr_time_from_sec( (tm_initial.tm_yday * SECONDS_PER_DAY) +
                                                     (tm_initial.tm_hour * SECONDS_PER_HOUR) +
                                                     (tm_initial.tm_min  * SECONDS_PER_MINUTE) +
                                                     (tm_initial.tm_sec  * 1) );
                    expected_mday = 1;
                    break;

                case INTERVAL_MONTH:
                    start_time -= apr_time_from_sec( ((tm_initial.tm_mday - 1) * SECONDS_PER_DAY) +
                                                      (tm_initial.tm_hour * SECONDS_PER_HOUR) +
                                                      (tm_initial.tm_min  * SECONDS_PER_MINUTE) +
                                                      (tm_initial.tm_sec  * 1) );
                    expected_mday = 1;
                    break;

                case INTERVAL_WEEK:
                    if ( weeks_start_on_mondays ) {
                        tm_initial.tm_wday = (6 + tm_initial.tm_wday) % 7;
                    }
                    start_time -= apr_time_from_sec( (tm_initial.tm_wday * SECONDS_PER_DAY) +
                                                     (tm_initial.tm_hour * SECONDS_PER_HOUR) +
                                                     (tm_initial.tm_min  * SECONDS_PER_MINUTE) +
                                                     (tm_initial.tm_sec  * 1) );
                    expected_mday = tm_initial.tm_mday;
                    break;

                case INTERVAL_DAY:
                    start_time -= apr_time_from_sec( (tm_initial.tm_hour * SECONDS_PER_HOUR) +
                                                     (tm_initial.tm_min  * SECONDS_PER_MINUTE) +
                                                     (tm_initial.tm_sec  * 1) );
                    expected_mday = tm_initial.tm_mday;
                    break;

                case INTERVAL_HOUR:
                case INTERVAL_MINUTE:
                case INTERVAL_SECOND:
                    break;

                default:
                    ap_log_perror(APLOG_MARK, APLOG_CRIT, 0, NULL, "unknown interval [%d]\n", interval);
                    start_time = 0;
                    break;
            }

            /*
             * If the time of day is not equal to midnight then we need to
             * adjust for daylight saving time.  Adjust the time backwards by
             * the value of the hour, minute and second fields.  If the day of
             * the month is not as expected one then we must have adjusted back
             * to the previous day so add 24 hours worth of seconds.
             */

            log_timeutils_explode_localtime(&tm_adjusted, start_time);

            if ((tm_adjusted.tm_hour != 0) || (tm_adjusted.tm_min != 0) || (tm_adjusted.tm_sec != 0)) {
                apr_time_t adjust = -1 * apr_time_from_sec( (tm_adjusted.tm_hour * SECONDS_PER_HOUR) + 
                                                            (tm_adjusted.tm_min  * SECONDS_PER_MINUTE) +
                                                            (tm_adjusted.tm_sec  * 1));
                char sign = '-';

                if (tm_adjusted.tm_mday != expected_mday) {
                    adjust += apr_time_from_sec( SECONDS_PER_DAY );
                    sign = '+';
                }
                start_time += adjust;

                if (adjust < 0) {
                    adjust = -adjust;
                }
                ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, NULL, "adjust for DST: %02d/%02d/%04d %02d:%02d:%02d -- %c%0d:%02d:%02d\n",
                                tm_initial.tm_mday,
                                tm_initial.tm_mon + 1,
                                tm_initial.tm_year + 1900,
                                tm_initial.tm_hour,
                                tm_initial.tm_min,
                                tm_initial.tm_sec,
                                sign,
                                (int)(apr_time_sec(adjust) / SECONDS_PER_HOUR),
                                (int)(apr_time_sec(adjust) / SECONDS_PER_HOUR % SECONDS_PER_HOUR),
                                (int)(apr_time_sec(adjust) % SECONDS_PER_HOUR)
                                );
            }
            break;

        case INTERVAL_HOUR:
            start_time -= apr_time_from_sec(tm_initial.tm_sec + tm_initial.tm_min * SECONDS_PER_MINUTE);
            break;

        case INTERVAL_MINUTE:
            start_time -= apr_time_from_sec(tm_initial.tm_sec);
            break;

        case INTERVAL_SECOND:
            /* no adjustment needed */
            break;

        default:
            ap_log_perror(APLOG_MARK, APLOG_CRIT, 0, NULL, "unknown interval [%d]\n", interval);
            start_time = 0;
            break;
    }

    return start_time;
}


/*
 * log_timeutils_start_of_next_interval
 *
 * To determine the time of the start of the next interval add just enough to
 * move beyond the start of the next interval and then determine the time of
 * the start of that interval.
 *
 * There is a potential problem if the start or end of daylight saving time
 * occurs during the current interval.
 *
 * @param       apr_time_t
 * @param       interval_t
 * @returns     apr_time_t
 */

apr_time_t log_timeutils_start_of_next_interval(apr_time_t time_now, interval_t interval) {

    apr_time_t start_time = 0;

    switch (interval) {
        case INTERVAL_INFINITY:
            break;

        case INTERVAL_YEAR:
            start_time = (time_now + apr_time_from_sec((366 * SECONDS_PER_DAY) + DST_ALLOWANCE));
            break;

        case INTERVAL_MONTH:
            start_time = (time_now + apr_time_from_sec((31 * SECONDS_PER_DAY) + DST_ALLOWANCE));
            break;

        case INTERVAL_WEEK:
            start_time = (time_now + apr_time_from_sec(SECONDS_PER_WEEK + DST_ALLOWANCE));
            break;

        case INTERVAL_DAY:
            start_time = (time_now + apr_time_from_sec(SECONDS_PER_DAY + DST_ALLOWANCE));
            break;

        case INTERVAL_HOUR:
            start_time = time_now + apr_time_from_sec(SECONDS_PER_HOUR + LEAP_SECOND_ALLOWANCE);
            break;

        case INTERVAL_MINUTE:
            start_time = time_now + apr_time_from_sec(SECONDS_PER_MINUTE + LEAP_SECOND_ALLOWANCE);
            break;

        case INTERVAL_SECOND:
            start_time = time_now + apr_time_from_sec(1);
            break;

        default:
            ap_log_perror(APLOG_MARK, APLOG_INFO, 0, NULL, "unknown interval [%d]\n", interval);
            break;
    }

    return log_timeutils_start_of_current_interval(start_time, interval);
}


/*
 * log_timeutils_interval_as_text
 *
 * @param       interval_t
 * @returns     const char*
 */
const char *log_timeutils_interval_as_text(interval_t interval) {
    return interval_names[interval];
}

/* ---- end time functions ---- */




/* ---- begin debugging functions ---- */

/*
 * log_timeutils_timestamp_as_text
 *
 * @param       apr_time_t
 * @returns     char*
 */
char *log_timeutils_timestamp_as_text(apr_time_t t) {

    static char     buffer[APR_PATH_MAX];
    apr_time_exp_t  tm;
    apr_size_t      rs;

    log_timeutils_explode_localtime(&tm, t);
    apr_strftime(buffer, &rs, APR_PATH_MAX, "%Y/%m/%d-%H:%M:%S %Z", &tm);

    return buffer;
}


/*
 * log_timeutils_get_version
 *
 * @returns     const char*
 */
const char *log_timeutils_get_version(void) {
    return RCS_ID;
}

/* ---- end debugging functions ---- */


/* ====================================================================
 * Copyright (c) 1995-1999 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */


/*
 * Copyright (c) 1996-1999 by Ford & Mason Ltd
 *
 * This software was submitted by Ford & Mason Ltd to the Apache Software
 * Foundation in December 1999.  Future revisions and derivatives of this
 * source code must acknowledge Ford & Mason Ltd as the original contributor of
 * this module.  All other licensing and usage conditions are those of the
 * Apache Software Foundation.
 *
 * Originally written by Andrew Ford <A.Ford@ford-mason.co.uk>
 *
 */
